#include <iostream>
#include <fstream>
#include <cmath>

#include "dff.h"
using namespace std;

namespace Rw {

uint32 oc, fc, gc, mc; /* object, frame, geometry, and material counter */

void CDff::CalcNormal(Vector3f &n, Vector3f &v0, Vector3f &v1, Vector3f &v2)
{
        float a[3], b[3];
        float len;

        /* Calculate the normal */
        a[0] = v1[0]-v0[0];
        a[1] = v1[1]-v0[1];
        a[2] = v1[2]-v0[2];
        b[0] = v2[0]-v0[0];
        b[1] = v2[1]-v0[1];
        b[2] = v2[2]-v0[2];
        n[0] = a[1]*b[2] - a[2]*b[1];
        n[1] = a[2]*b[0] - a[0]*b[2];
        n[2] = a[0]*b[1] - a[1]*b[0];

        /* Make it 1 unit long */
        len = sqrt(n[0] * n[0] +
                   n[1] * n[1] +
                   n[2] * n[2]);
        if (len == 0)
                len = 1;
        n[0] /= len;
        n[1] /= len;
        n[2] /= len;
}

void CDff::MakeNormals(RpGeometry &g, RpMesh &s)
{
	uint32 i;

	/* Strip */
	if (g.FaceType == 1) {
		s.Normals = new Vector3f[s.numIndices-2];
		for (i = 0; i < s.numIndices-2; i++) {
			CalcNormal(s.Normals[i], g.Vertex[s.Indices[i]],
			  g.Vertex[s.Indices[i+1+(i % 2)]],
			  g.Vertex[s.Indices[i+2-(i % 2)]]);
		}
	/* List */
	} else {
		s.Normals = new Vector3f[s.numIndices/3];
		for (i = 0; i < s.numIndices-2; i += 3) {
			CalcNormal(s.Normals[i/3], g.Vertex[s.Indices[i]],
			  g.Vertex[s.Indices[i+1]],
			  g.Vertex[s.Indices[i+2]]);
		}
	}
}

void CDff::ReadFromFile(istream &dff)
{
	uint32 size;

	dff.seekg(0, ios::end);
	size = dff.tellg();
	dff.seekg(0, ios::beg);
	ReadFromFile(dff, 0, size);
}

void CDff::ReadFromFile(istream &dff, uint32 start, uint32 end)
{
	CHeader rwh;

	if ((uint32) dff.tellg() < end) {
		rwh.Read(dff);
		if (rwh.GetType() != rwID_CLUMP)
			return;
		Clp = new RpClump;
		ParseSection(*Clp, rwh, 0, dff);
		numClumps++;
	}
}

void CDff::ParseSection(RpClump &c, CHeader &rwh, uint32 parent, istream &dff)
{
	uint32 i;
	RpGeometry *geo;

	switch (rwh.GetType()) {
	case rwID_CLUMP:
		oc = fc = gc = 0;
		c.numAtomics = 0;
		c.numFrames = 0;
		c.numGeometries = 0;
		c.Atm = NULL;
		c.Frm = NULL;
		c.Geo = NULL;
		ReadSection(c, rwh, dff);
		break;
	case rwID_STRUCT:
		ReadStruct(c, rwh, parent, dff);
		break;
	case rwID_FRAME:
		ReadFrame(c.Frm[fc], rwh.GetSize(), dff);
		fc++;
		break;
	case rwID_GEOMETRY:
		ReadSection(c, rwh, dff);
		gc++;
		break;
	case rwID_MATERIAL:
		ReadSection(c, rwh, dff);
		mc++;
		break;
	case rwID_ATOMIC:
		/* It is necessary to read all atomics at once.
		 * The clump size of gta3 ps2 dffs is often not correct and we
		 * would miss two atomics if we read them when we see them */
		for (i = 0; i < c.numAtomics; i++) {
			ReadSection(c, rwh, dff);
			rwh.Read(dff);
			oc++;
		}
		break;
	case rwID_BINMESH:
		geo = &c.Geo[gc];

		geo->FaceType = ReadDword(dff);
		geo->numSplits = ReadDword(dff);
		geo->numIndices = ReadDword(dff);

		geo->Split = new RpMesh[geo->numSplits];

		for (i = 0; i < geo->numSplits; i++) {
			RpMesh *s;

			s = &geo->Split[i];

			s->numIndices = ReadDword(dff);
			s->MatIndex = ReadDword(dff);
			if (!geo->isPs2) {
				s->Indices = new uint32[s->numIndices];
				dff.read((char*)s->Indices,
				  sizeof(uint32) * s->numIndices);
				if (!(geo->Flags & rpGEOMETRYNORMALS))
					MakeNormals(*geo, *s);
			}
		}
		break;
	case rwID_MATERIALEFFECTS:
		ReadMatfx(c, rwh, parent, dff);
		break;
	case rwID_SPECULARMAT:
		c.Geo[gc].Mat[mc].SpecLevel = ReadFloat(dff);
		c.Geo[gc].Mat[mc].SpecName = new char[rwh.GetSize()-8];
		dff.read(c.Geo[gc].Mat[mc].SpecName, rwh.GetSize()-8);
		dff.seekg(4, ios::cur);
		break;
	case rwID_REFLECTIONMAT:
		c.Geo[gc].Mat[mc].RefImageAmount[0] = ReadFloat(dff);
		c.Geo[gc].Mat[mc].RefImageAmount[1] = ReadFloat(dff);
		c.Geo[gc].Mat[mc].RefImageAmount[2] = ReadFloat(dff);
		c.Geo[gc].Mat[mc].RefImageAmount[3] = ReadFloat(dff);
		c.Geo[gc].Mat[mc].RefIntensity = ReadFloat(dff);
		dff.seekg(4, ios::cur);
		break;
	case rwID_MATLIST:
	case rwID_EXTENSION:
	case rwID_FRAMELIST:
	case rwID_GEOMETRYLIST:
	case rwID_TEXTURE:
		ReadSection(c, rwh, dff);
		break;
	case rwID_NATIVEDATA:
		/*
		 * Actually there is no such thing as SA compression or
		 * optimization. SA just uses by convention a little
		 * different format that could probably be also read by
		 * Vice City (it uses it in fact for skinning if I'm not
		 * mistaken).
		 * SA can read VC dffs (weapons in SA use that format)
		 */
		/* These should be combined when I have uncovered the secrets
		 * of PS2 optimization */
		if (IsSaCompression(c.Geo[gc], dff))
			ReadNativeDataSaPs2(c.Geo[gc], dff);
		else
			ReadNativeDataVcPs2(c.Geo[gc], dff);
		break;
	default:
		dff.seekg(rwh.GetSize(), ios::cur);
		break;
	}
}

void CDff::ReadSection(RpClump &c, CHeader &rwh, istream &dff)
{
	CHeader child;
	uint32 end;

	end = rwh.GetSize() + dff.tellg();
	while ((uint32) dff.tellg() < end) {
		child.Read(dff);
		ParseSection(c, child, rwh.GetType(), dff);
	}
}

void CDff::ReadFrame(RwFrame &f, uint32 size, istream &dff)
{
	f.Name = new char[size+1];
	dff.read(f.Name, size);
	f.Name[size] = '\0';
}

void CDff::ReadMatfx(RpClump &c, CHeader &rwh, uint32 parent, istream &dff)
{
	int type;
	CHeader texh;

	type = ReadDword(dff);
	if (type != 2)
		return;
	dff.seekg(16, ios::cur);

	type = PeekDword(dff);
	if (type == 0) {
		dff.seekg(4, ios::cur);
		return;
	}
	
	c.Geo[gc].Mat[mc].HasRef = 1;

	texh.Read(dff);
	ReadSection(c, texh, dff);

	dff.seekg(4, ios::cur);
}

void CDff::ReadStruct(RpClump &c, CHeader &rwh, uint32 parent, istream &dff)
{
	uint32 i;
	RpGeometry *g;
	RpMaterial *m;
	RpTexture *t;

	switch (parent) {
	case rwID_CLUMP:
		c.numAtomics = ReadDword(dff);
		if (rwh.GetVersion() != GTA3_1 && rwh.GetVersion() != GTA3_2) {
			c.numLights = ReadDword(dff);
			c.numCameras = ReadDword(dff);
		}
		c.Atm = new RpAtomic[c.numAtomics];
		break;
	case rwID_FRAMELIST:
		c.numFrames = ReadDword(dff);
		c.Frm = new RwFrame[c.numFrames];
		for (i = 0; i < c.numFrames; i++) {
			RwFrame *f;

			f = &c.Frm[i];

			dff.read((char*)f->Rotation, sizeof(float)*9);
			dff.read((char*)f->Position, sizeof(float)*3);
			f->Parent = ReadDword(dff);
			f->Unknown = ReadDword(dff);
			f->Name = NULL;
		}
		break;
	case rwID_GEOMETRYLIST:
		c.numGeometries = ReadDword(dff);
		c.Geo = new RpGeometry[c.numGeometries];
		break;
	case rwID_GEOMETRY:
		g = &c.Geo[gc];
		mc = 0;

		g->Flags = ReadWord(dff);
		g->numUVs = ReadByte(dff);
		g->isPs2 = ReadByte(dff);
		g->numTriangles = ReadDword(dff);
		g->numVertices = ReadDword(dff);
		g->numMorphTargets = ReadDword(dff);
		if (rwh.GetVersion() == GTA3_1 || rwh.GetVersion() == GTA3_2 ||
		    rwh.GetVersion() == VCPS2) {
			dff.seekg(12, ios::cur);
		}
		g->VColor = NULL;
		g->UV = NULL;
		g->UV2 = NULL;
		g->Face = NULL;
		g->Vertex = NULL;
		g->Normal = NULL;

		if (!g->isPs2) {
			if ((g->Flags & rpGEOMETRYPRELIT) != 0) {
				g->VColor = new Color4b[g->numVertices];
				dff.read((char*)g->VColor, 4*1*g->numVertices);
			}
			if ((g->Flags & rpGEOMETRYTEXTURED) != 0) {
				g->UV = new Vector2f[g->numVertices];
				dff.read((char*)g->UV, 2*4*g->numVertices);
			}
			if ((g->Flags & rpGEOMETRYTEXTURED2) != 0) {
				g->UV = new Vector2f[g->numVertices];
				dff.read((char*)g->UV, 2*4*g->numVertices);
				g->UV2 = new Vector2f[g->numVertices];
				dff.read((char*)g->UV2, 2*4*g->numVertices);
			}
			g->Face = new Face4i[g->numTriangles];
			dff.read((char*)g->Face, 4*2*g->numTriangles);
		}

		dff.read((char*)g->Bs, 4*4);
		g->BsPos = ReadDword(dff);
		g->BsNor = ReadDword(dff);

		if (!g->isPs2) {
			g->Vertex = new Vector3f[g->numVertices];
			dff.read((char*)g->Vertex, 3*4*g->numVertices);

			if ((g->Flags & rpGEOMETRYNORMALS) != 0) {
				g->Normal = new Vector3f[g->numVertices];
				dff.read((char*)g->Normal, 3*4*g->numVertices);
			}
		}
		break;
	case rwID_MATLIST:
		c.Geo[gc].numMaterials = ReadDword(dff);
		c.Geo[gc].Mat = new RpMaterial[c.Geo[gc].numMaterials];
		dff.seekg(4*c.Geo[gc].numMaterials, ios::cur);
		break;
	case rwID_MATERIAL:
		m = &c.Geo[gc].Mat[mc];

		m->Flags = ReadDword(dff);
		dff.read((char*)m->Color, 4);
		m->Unused = ReadDword(dff);
		m->HasTex = ReadDword(dff);
		m->SurfaceProps[0] = ReadFloat(dff);
		m->SurfaceProps[1] = ReadFloat(dff);
		m->SurfaceProps[2] = ReadFloat(dff);
		m->TexRead = 0;
		m->HasRef = 0;
		m->SpecName = NULL;
		break;
	case rwID_TEXTURE:
		m = &c.Geo[gc].Mat[mc];
		t = &m->Tex;

		/* If the texture has already been read, or there has been
		 * no texture, read reflection map. */
		if (m->TexRead || !m->HasTex)
			t = &m->Reflection;

		t->FilterFlags = ReadWord(dff);
		t->Unknown = ReadWord(dff);

		/* texture name */
		rwh.Read(dff);
		t->Tex = new char[rwh.GetSize()+1];
		dff.read(t->Tex, rwh.GetSize());
		t->Tex[rwh.GetSize()] = '\0';

		/* alpha name */
		rwh.Read(dff);
		t->Alpha = new char[rwh.GetSize()+1];
		dff.read(t->Alpha, rwh.GetSize());
		t->Alpha[rwh.GetSize()] = '\0';

		m->TexRead = 1;
		break;
	case rwID_ATOMIC:
		c.Atm[oc].FrmIndex = ReadDword(dff);
		c.Atm[oc].GeoIndex = ReadDword(dff);
		c.Atm[oc].Unknown0 = ReadDword(dff);
		c.Atm[oc].Unknown1 = ReadDword(dff);
		break;
	default:
		dff.seekg(rwh.GetSize(), ios::cur);
	}
}

float CDff::GetSaVertScale(RpGeometry &g)
{
	/* This seems to work but is probably only coincidence*/
	if (g.Flags & rpGEOMETRYNORMALS)
		return SAVERTSCALE2;
	else
		return SAVERTSCALE1;
}

int CDff::IsSameVert(float *v0, float *v1)
{
	if (v0[0] == v1[0] && v0[1] == v1[1] &&
	  v0[2] == v1[2])
		return 1;
	else
		return 0;
}

void CDff::Ps2ToPc(RpGeometry &g, struct ps2split *splits)
{
	uint32 i, j, k;
	int faceinc;

	/* Correct index count for each split */
	for (i = 0; i < g.numSplits; i++)
		g.Split[i].numIndices = splits[i].vertexc+splits[i].extraverts;

	/* Correct index count for the whole mesh */
	g.numIndices = 0;
	for (i = 0; i < g.numSplits; i++)
		g.numIndices += g.Split[i].numIndices;

	/* Calculate vertex count */
	g.numVertices = 0;
	for (i = 0; i < g.numSplits; i++)
		g.numVertices += g.Split[i].numIndices;

	/* Write Vertices into Geometry struct */
	g.Vertex = new Vector3f[g.numVertices];
	for (i = k = 0; i < g.numSplits; i++) {
		for (j = 0; j < splits[i].vertexc; j++) {
			g.Vertex[j+k][0] = splits[i].vertices[j][0];
			g.Vertex[j+k][1] = splits[i].vertices[j][1];
			g.Vertex[j+k][2] = splits[i].vertices[j][2];
		}
		k += j;
	}

	/* Write UV Coordinates into Geometry Struct if we have any */
	if (g.Flags & rpGEOMETRYTEXTURED) {
		g.UV = new Vector2f[g.numVertices];
		for (i = k = 0; i < g.numSplits; i++) {
			for (j = 0; j < splits[i].vertexc; j++) {
				g.UV[j+k][0] = splits[i].uv[j][0];
				g.UV[j+k][1] = splits[i].uv[j][1];
			}
			k += j;
		}
	} else if (g.Flags & rpGEOMETRYTEXTURED2) {
		g.UV = new Vector2f[g.numVertices];
		g.UV2 = new Vector2f[g.numVertices];
		for (i = k = 0; i < g.numSplits; i++) {
			for (j = 0; j < splits[i].vertexc; j++) {
				g.UV[j+k][0] = splits[i].uv[j][0];
				g.UV[j+k][1] = splits[i].uv[j][1];
			}
			for (j = 0; j < splits[i].vertexc; j++) {
				g.UV2[j+k][0] = splits[i].uv[j][0];
				g.UV2[j+k][1] = splits[i].uv[j][1];
			}
			k += j;
		}
	}

	/* Write Vertex colors into Geometry Struct if we have any */
	if (g.Flags & rpGEOMETRYPRELIT) {
		g.VColor = new Color4b[g.numVertices];
		for (i = k = 0; i < g.numSplits; i++) {
			for (j = 0; j < splits[i].vertexc; j++) {
				g.VColor[j+k][0] = splits[i].vc[j][0];
				g.VColor[j+k][1] = splits[i].vc[j][1];
				g.VColor[j+k][2] = splits[i].vc[j][2];
				g.VColor[j+k][3] = splits[i].vc[j][3];
			}
			k += j;
		}
	}

	/* Write Normals into Geometry struct if we have any */
	k = 0;
	if (g.Flags & rpGEOMETRYNORMALS) {
		g.Normal = new Vector3f[g.numVertices];
		for (i = k = 0; i < g.numSplits; i++) {
			for (j = 0; j < splits[i].vertexc; j++) {
				g.Normal[j+k][0] = splits[i].normals[j][0];
				g.Normal[j+k][1] = splits[i].normals[j][1];
				g.Normal[j+k][2] = splits[i].normals[j][2];
			}
			k += j;
		}
	}

	/* Write an index list */
	faceinc = 0;
	for (i = 0; i < g.numSplits; i++) {
		uint32 k;
		g.Split[i].Indices = new uint32[g.Split[i].numIndices];
		for (j = k = 0; j < splits[i].vertexc; j++, k++) {
			g.Split[i].Indices[k] = j+faceinc;
			if (splits[i].flags && splits[i].flags[j+1] == 0x8000) {
				k++;
				g.Split[i].Indices[k] = j+faceinc;
			}
		}
		faceinc += j;
	}

	/* Write Faces. Note: This is not elegant, but useless anyhow */
	if (g.FaceType == 1) {
		g.numTriangles = g.numIndices - 2*g.numSplits;
		g.Face = new Face4i[g.numTriangles];
		faceinc = 0;
		for (i = 0; i < g.numSplits; i++) {
			for (j = 0; j < g.Split[i].numIndices-2; j++) {
				g.Face[j+faceinc][0] = g.Split[i].Indices[j];
				g.Face[j+faceinc][1] = g.Split[i].Indices[j+1];
				g.Face[j+faceinc][2] = 0;
				g.Face[j+faceinc][3] = g.Split[i].Indices[j+2];
			}
			faceinc += j;
		}
	} else {
		g.numTriangles = g.Split[i].numIndices/3;
		cout << "TRILIST not supported\n";
		exit(1);
	}

	/* Free everything */
	for (i = 0; i < g.numSplits; i++) {
		for (j = 0; j < splits[i].vertexc; j++) {
			free(splits[i].vertices[j]);
			free(splits[i].uv[j]);
			free(splits[i].uv2[j]);
			free(splits[i].vc[j]);
			free(splits[i].normals[j]);
		}
		free(splits[i].vertices);
		free(splits[i].uv);
		free(splits[i].uv2);
		free(splits[i].vc);
		free(splits[i].normals);
		free(splits[i].flags);
	}
	for (i = 0; i < g.numSplits; i++)
		if (!(g.Flags & rpGEOMETRYNORMALS))
			MakeNormals(g, g.Split[i]);
}

void CDff::ReadNativeDataVcPs2(RpGeometry &g, istream &dff)
{
	uint32 i, k;
	uint32 blockvertcount, oldvertcount;
	int isnotactor, islast;
	int subsecthdr;
	int splitsize, nextsplit;
	signed char normal;
	struct ps2split *splits;
	struct ps2split *cursplit;


	splits = (struct ps2split*) malloc(g.numSplits*sizeof(struct ps2split));
	/* Skip header and platform */
	dff.seekg(16, ios::cur);

	for (i = 0; i < g.numSplits; i++) {
		cursplit = &splits[i];

		splitsize = ReadDword(dff);
		nextsplit = splitsize + 4 + dff.tellg();

		isnotactor = ReadDword(dff);
		if (isnotactor)
			dff.seekg(16, ios::cur);

		cursplit->vertices = NULL;
		cursplit->uv = NULL;
		cursplit->uv2 = NULL;
		cursplit->vc = NULL;
		cursplit->normals = NULL;

		/* Loop through blocks */
		cursplit->vertexc = 0;
		islast = 0;
		while (!islast) {
			if (!isnotactor)
				dff.seekg(48, ios::cur);

			/* Vertices */
			dff.seekg(12, ios::cur);
			subsecthdr = ReadDword(dff);
			blockvertcount = (subsecthdr & 0x00ff0000) >> 16;
			oldvertcount = cursplit->vertexc;
			cursplit->vertexc += blockvertcount;

			cursplit->vertices = (float **)
			  realloc(cursplit->vertices, cursplit->vertexc*4);
			cursplit->flags = NULL;
			cursplit->extraverts = 0;
			for (k = oldvertcount; k < cursplit->vertexc; k++) {
				cursplit->vertices[k] = (float *) malloc(3*4);
				cursplit->vertices[k][0] = ReadFloat(dff);
				cursplit->vertices[k][1] = ReadFloat(dff);
				cursplit->vertices[k][2] = ReadFloat(dff);
			}
			/* Skip padding */
			if ((blockvertcount*12 % 16) != 0)
				dff.seekg(16-(blockvertcount*12%16), ios::cur);


			/* UV Coordinates */
			dff.seekg(16, ios::cur);
			cursplit->uv = (float **)
			  realloc(cursplit->uv, cursplit->vertexc*4);
			cursplit->uv2 = (float **)
			  realloc(cursplit->uv2, cursplit->vertexc*4);
			for (k = oldvertcount; k < cursplit->vertexc; k++) {
				cursplit->uv[k] = (float *) malloc(2*4);
				cursplit->uv2[k] = (float *) malloc(2*4);
				cursplit->uv[k][0] = ReadFloat(dff);
				cursplit->uv[k][1] = ReadFloat(dff);
			}
			/* Skip padding */
			if ((blockvertcount*8 % 16) != 0)
				dff.seekg(16-(blockvertcount*8%16), ios::cur);


			/* Vertex colors */
			dff.seekg(16, ios::cur);
			cursplit->vc = (byte **)
			  realloc(cursplit->vc, cursplit->vertexc*4);
			for (k = oldvertcount; k < cursplit->vertexc; k++) {
				cursplit->vc[k] = (byte *) malloc(1*4);
				cursplit->vc[k][0] = ReadByte(dff);
				cursplit->vc[k][1] = ReadByte(dff);
				cursplit->vc[k][2] = ReadByte(dff);
				cursplit->vc[k][3] = ReadByte(dff);
			}
			/* Skip padding */
			if ((blockvertcount*4 % 16) != 0)
				dff.seekg(16-(blockvertcount*4%16), ios::cur);


			/* Normals */
			dff.seekg(16, ios::cur);
			cursplit->normals = (float **)
			  realloc(cursplit->normals, cursplit->vertexc*4);
			for (k = oldvertcount; k < cursplit->vertexc; k++) {
				cursplit->normals[k] = (float *) malloc(3*4);
				normal = ReadByte(dff);
				cursplit->normals[k][0] = normal * VCNORMSCALE;
				normal = ReadByte(dff);
				cursplit->normals[k][1] = normal * VCNORMSCALE;
				normal = ReadByte(dff);
				cursplit->normals[k][2] = normal * VCNORMSCALE;
			}
			/* Skip padding */
			if ((blockvertcount*3 % 16) != 0)
				dff.seekg(16-(blockvertcount*3%16), ios::cur);

			/* Skip end of block and see whether it was the last */
			dff.seekg(8, ios::cur);
			islast = ReadDword(dff);
			if (islast == 0x11000000)
				islast = 1;
			else
				islast = 0;
			dff.seekg(4, ios::cur);
		}
		/* Jump to next split and skip unknowns (probably skin data) */
		dff.seekg(nextsplit);
	}
	Ps2ToPc(g, splits);
	free(splits);
}

void CDff::ReadNativeDataSaPs2(RpGeometry &g, istream &dff)
{
	uint32 i, j;
	uint32 splitsize, nextsplit;
	uint32 vertexpos, uvpos, normalpos, vcpos;
	uint32 startaddress;
	float vertscale;
	int16 tmp;
	struct ps2split *splits;
	struct ps2split *cursplit;

	splits = (struct ps2split*) malloc(g.numSplits*sizeof(struct ps2split));
	/* Skip header and platform */
	dff.seekg(16, ios::cur);

	vertscale = GetSaVertScale(g);

	for (i = 0; i < g.numSplits; i++) {
		cursplit = &splits[i];

		splitsize = ReadDword(dff);
		nextsplit = splitsize + 4 + dff.tellg();

		dff.seekg(4, ios::cur);
		startaddress = dff.tellg();

		dff.seekg(4, ios::cur);
		vertexpos = ReadDword(dff);
		vertexpos = vertexpos*0x10 + startaddress; 
		if ((g.Flags & rpGEOMETRYTEXTURED) ||
		    (g.Flags & rpGEOMETRYTEXTURED2)) {
			dff.seekg(28, ios::cur);
			uvpos = ReadDword(dff);
			uvpos = uvpos*0x10 + startaddress; 
		}
		if (g.Flags & rpGEOMETRYPRELIT) {
			dff.seekg(28, ios::cur);
			vcpos = ReadDword(dff);
			vcpos = vcpos*0x10 + startaddress; 
		}
		if (g.Flags & rpGEOMETRYNORMALS) {
			dff.seekg(28, ios::cur);
			normalpos = ReadDword(dff);
			normalpos = normalpos*0x10 + startaddress; 
		}

		cursplit->vertexc = g.Split[i].numIndices;

		cursplit->vertices = (float **)
		  malloc(cursplit->vertexc*4);
		cursplit->flags = (uint16 *)
		  malloc(cursplit->vertexc*sizeof(uint16));
		cursplit->uv = (float **)
		  malloc(cursplit->vertexc*4);
		cursplit->uv2 = (float **)
		  malloc(cursplit->vertexc*4);
		cursplit->vc = (byte **)
		  malloc(cursplit->vertexc*4);
		cursplit->normals = (float **)
		  malloc(cursplit->vertexc*4);

		/* Read Vertices */
		dff.seekg(vertexpos);
		cursplit->extraverts = 0;
		/* How can I determine the correct scaling factor?
		 * Vehicles use 1/1024 every other object uses 1/128 */
		for (j = 0; j < cursplit->vertexc; j++) {
			cursplit->vertices[j] = (float *) malloc(3*4);
			tmp = ReadWord(dff);
			cursplit->vertices[j][0] = tmp * vertscale;
			tmp = ReadWord(dff);
			cursplit->vertices[j][1] = tmp * vertscale;
			tmp = ReadWord(dff);
			cursplit->vertices[j][2] = tmp * vertscale;
			if ((cursplit->flags[j] = ReadWord(dff)) == 0x8000)
				cursplit->extraverts++;
		}

		/* Read UV coordinates */
		if (g.Flags & rpGEOMETRYTEXTURED) {
			dff.seekg(uvpos);
			for (j = 0; j < cursplit->vertexc; j++) {
				cursplit->uv[j] = (float *) malloc(2*4);
				tmp = ReadWord(dff);
				cursplit->uv[j][0] = tmp * SAUVSCALE;
				tmp = ReadWord(dff);
				cursplit->uv[j][1] = tmp * SAUVSCALE;
			}
			for (j = 0; j < cursplit->vertexc; j++)
				cursplit->uv2[j] = (float *) malloc(2*4);
		} else if (g.Flags & rpGEOMETRYTEXTURED2) {
			/* Note that this is incorrect */	
			dff.seekg(uvpos);
			for (j = 0; j < cursplit->vertexc; j++) {
				cursplit->uv[j] = (float *) malloc(2*4);
				cursplit->uv2[j] = (float *) malloc(2*4);

				tmp = ReadWord(dff);
				cursplit->uv[j][0] = tmp * SAUVSCALE;
				tmp = ReadWord(dff);
				cursplit->uv[j][1] = tmp * SAUVSCALE;

				tmp = ReadWord(dff);
				cursplit->uv2[j][0] = tmp * SAUVSCALE;
				tmp = ReadWord(dff);
				cursplit->uv2[j][1] = tmp * SAUVSCALE;
			}
		} else {
			for (j = 0; j < cursplit->vertexc; j++) {
				cursplit->uv[j] = (float *) malloc(2*4);
				cursplit->uv2[j] = (float *) malloc(2*4);
			}
		}

		/* Read Vertex colors */
		if (g.Flags & rpGEOMETRYPRELIT) {
			dff.seekg(vcpos);
			for (j = 0; j < cursplit->vertexc; j++) {
				cursplit->vc[j] = (byte *) malloc(4*1);
				cursplit->vc[j][0] = ReadByte(dff);
				cursplit->vc[j][1] = ReadByte(dff);
				cursplit->vc[j][2] = ReadByte(dff);
				cursplit->vc[j][3] = ReadByte(dff);
			}
		} else {
			for (j = 0; j < cursplit->vertexc; j++)
				cursplit->vc[j] = (byte *) malloc(4*1);
		}

		/* Read Normals */
		if (g.Flags & rpGEOMETRYNORMALS) {
			dff.seekg(normalpos);
			for (j = 0; j < cursplit->vertexc; j++) {
				cursplit->normals[j] = (float *) malloc(3*4);
				tmp = ReadByte(dff);
				cursplit->normals[j][0] = tmp * SANORMSCALE;
				tmp = ReadByte(dff);
				cursplit->normals[j][1] = tmp * SANORMSCALE;
				tmp = ReadByte(dff);
				cursplit->normals[j][2] = tmp * SANORMSCALE;
				ReadByte(dff);
			}
		} else {
			for (j = 0; j < cursplit->vertexc; j++)
				cursplit->normals[j] = (float *) malloc(3*4);
		}

		dff.seekg(nextsplit);
	}
	Ps2ToPc(g, splits);
	free(splits);
}

int CDff::IsSaCompression(RpGeometry &g, istream &dff)
{
	int isnotactor;
	CHeader rwh;

	rwh.Read(dff);
	dff.seekg(8, ios::cur);
	isnotactor = ReadDword(dff);
	dff.seekg(-24, ios::cur);
	if ((rwh.GetVersion() == SA) && (isnotactor == 0))
		return 1;
	else
		return 0;
}


CDff::CDff()
{
	Clp = NULL;
	numClumps = 0;
}

CDff::~CDff()
{
	uint32 j, k;
	RpGeometry *g;
	RpMaterial *m;
	RpMesh *s;

	if (Clp == NULL)
		return;
	for (j = 0; j < Clp->numFrames; j++)
		if (Clp->Frm[j].Name != NULL)
			delete[] Clp->Frm[j].Name;
	delete[] Clp->Frm;

	for (j = 0; j < Clp->numGeometries; j++) {
		g = &Clp->Geo[j];
		for (k = 0; k < g->numMaterials; k++) {
			m = &g->Mat[k];
			if (m->HasTex) {
				delete[] m->Tex.Tex;
				delete[] m->Tex.Alpha;
			}
			if (m->HasRef) {
				delete[] m->Reflection.Tex;
				delete[] m->Reflection.Alpha;
			}
			if (m->SpecName != NULL)
				delete[] m->SpecName;
		}
		delete[] g->Mat;

		delete[] g->Vertex;
		delete[] g->Face;
		if (g->Flags & rpGEOMETRYTEXTURED)
			delete[] g->UV;
		if (g->Flags & rpGEOMETRYTEXTURED2) {
			delete[] g->UV;
			delete[] g->UV2;
		}
		if (g->Flags & rpGEOMETRYPRELIT)
			delete[] g->VColor;
		if (g->Flags & rpGEOMETRYNORMALS)
			delete[] g->Normal;

		for (k = 0; k < g->numSplits; k++) {
			s = &g->Split[k];
			delete[] s->Indices;
			if ((g->Flags & rpGEOMETRYNORMALS) == 0)
				delete[] s->Normals;
		}
		delete[] g->Split;
	}
	delete[] Clp->Geo;

	delete[] Clp->Atm;

	delete Clp;
}

}
