#ifdef WIN32
  #include <windows.h>
  #include <direct.h>
  #include <GL/glew.h>
  #define _USE_MATH_DEFINES
  #define _CRT_SECURE_NO_DEPRECATE
#endif

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>
#include <iostream>
#include <fstream>
#include <vector>
#include <GL/gl.h>
#include <GL/glu.h>

#include <rwcore.h>
#include <rpworld.h>

#ifndef DFFVIEWER_H
#define DFFVIEWER_H

#define GAMEPC  0x1
#define GAMEPS2 0x2
#define GAMEIII	0x4
#define GAMEVICECITY 0x8
#define GAMESANANDREAS 0x10

using namespace std;

typedef unsigned int uint;
typedef unsigned char byte;
typedef byte color[4];

class CCamera {
private:
	GLfloat Position[3];
	GLfloat LookVector[3];
	GLfloat Yaw;
	GLfloat Pitch;
	GLfloat CamPositions[12][5];
	GLfloat FOV;
	GLfloat Frustum[6][4];
	GLfloat EyeDist;
	void CalcLookVector(void);
public:
	void SetYaw(GLfloat);
	GLfloat GetYaw(void);
	void SetPitch(GLfloat);
	GLfloat GetPitch(void);
	void SetPosition(GLfloat, GLfloat, GLfloat);
	void TurnLeftRight(GLfloat);
	void TurnUpDown(GLfloat);
	void MoveForwardBackward(GLfloat);
	void MoveUpDown(GLfloat);
	void StrafeLeftRight(GLfloat);
	void ChangeFOV(GLfloat);
	GLfloat GetFOV(void);
	void IncreaseEyeDist(GLfloat);
	void DecreaseEyeDist(GLfloat);
	void UpdateFrustum(void);
	bool isPointInFrustum(const GLfloat *);
	bool isTranslatedPointInFrustum(const GLfloat *);
	void Look(void);
	void LookLeftEye(void);
	void LookRightEye(void);
	float DistanceTo(float *vector);
	void ReadSavedPositions(void);
	void SavePosition(int PosNumber);
	void LoadPosition(int PosNumber);
	void GetPosition(GLfloat *x, GLfloat *y, GLfloat *z);
	CCamera();
	~CCamera();
};

typedef struct DirEntry DirEntry;
struct DirEntry {
	char FileName[48]; /* file name used by GTA */
	uint Start;
	uint Size;
	char ContFile[48]; /* the actual file */
	bool Unloadable;

	/* Either Dff or Txd */
	void *Data;
	DirEntry *Parent;
};

class CDirectory {
private:
	vector<DirEntry> FileList;
public:
	void AppendFromDir(FILE *Dir, const char *ContFile);
	void Append(DirEntry&);
	void Append(const char *FileName, uint Start, uint Size,
	            const char *ContFile, void * Data);
	DirEntry *GetEntry(const char *FileName);
	DirEntry *GetEntry(const char *Name, const char *Ext);
	void Dump(void);
};

/* Only type 0 (light) at the moment */
class C2dfx {
public:
	int ID;
	Vector3f PositionV;
	color Color;
	int Type;	/* Only 0 supported */

	/* Light */
	char CoronaTexture[24];
	char ShadowTexture[24];
	int DrawDist;
	int outer_range;
	int lamp_size;
	int inner_range;
	int corona_size;
	int control;
	int reflection_on_wet_roads;
	int lensflare;
	int dust;
};

class CWorldObj {
public:
	int ID;
	char ModelName[25];
	char TextureName[25];
	char AnimName[25];
	int ObjCount;
	float *DrawDist;
	int Flags;
	int TimeOn;
	int TimeOff;
	bool isTimed;

	RpClump *Clump;
	RwTexDictionary *TexDict;

	DirEntry *DffEntry;
	DirEntry *TxdEntry;

	int RefCount;

	void SetClump(RpClump *);
	RpClump *GetClump(void);
	void TransformFrame(RwFrame *, RwInt32);
	void DrawGeometry(RpGeometry&, bool);
	void Draw(bool DoTransformation, bool DrawTransp);
};

class CDefinitionList {
private:
	vector<CWorldObj> ObjList;
	vector<C2dfx> FxList;
public:
	void AppendFromFile(FILE *file);
	CWorldObj *GetObj(int ID);
};

typedef struct Instance Instance;
struct Instance {
	int ID;
	char Name[64];
	int Interior;
	Vector3f PositionV;
	Vector3f ScaleV;
	Vector4f Rotation;
	int LOD;

	CWorldObj *Obj;

	Instance *Parent;
	Instance *Child;
	int Index;

	bool isLOD;
	bool isIslandLOD;
};

class CInstanceList {
private:
	vector<Instance> InstList;

	int Index;
public:
	void AppendFromFile(FILE *file, char *name);
	void AppendFromPlainFile(FILE *file);
	void AppendFromBinFile(FILE *file);
	Instance *GetFirst(void);
	int GetLength(void);
	Instance *GetInst(int index);
	Instance *GetLOD(Instance *inst);
	Instance *GetCorrectInst(Instance *inst);
	void Dump(void);
};


class CFileLoader {
public:
	RpClump *LoadClump(DirEntry *);
	RwTexDictionary *LoadTexDict(DirEntry *);
	void FreeRwTextures(RwTexDictionary *);
	void LoadGLTextures(RwTexDictionary *);
	bool LoadDff(CWorldObj *obj);
	bool LoadTxd(CWorldObj *obj);
	void UnloadWorldObj(CWorldObj *obj);
	void LoadBinIpl(DirEntry *Entry);
};

class CWater {
private:
	byte indiceshi[128*128];
	byte indiceslo[64*64];
	float palette[48]; /* not sure */
	void DrawPatch(float x, float y, float patchsize, float height);
public:
	GLuint TexID;
	void Draw(void);
	void ReadWaterData(void);
};

typedef struct SWeather SWeather;
struct SWeather {
	color Amb;
	color Amb_Obj;
	color Amb_Bl;
	color Amb_Obj_bl;
	color Dir;
	color Sky_top;
	color Sky_bot;
	color SunCore;
	color SunCorona;
	float SunSz;
	float SprSz;
	float SprBght;
	int Shdw;
	int LightShd;
	int TreeShd;
	int PoleShd;
	float FarClp;
	float FogSt;
	float LightOnGround;
	color LowCloudsRGB;
	color TopCloudRGB;
	color BottomCloudRGB;

	/* GTA 3 only, multiplied with 0.6 (AK-73) */
	color Tint;

	color BlurRGB;
	color WaterRGBA;

	color RGBA1;
	color RGBA2;
	float CloudAlpha;
};

typedef struct SCloud SCloud;
struct SCloud {
	GLfloat Position[3];
	char TexName[12];
};

class CSky {
private:
	vector<SWeather> WeatherList;
	int numWeather;
	int WeatherSelector;
	GLfloat SunPos[3];
	SCloud BottomClouds[10];
	SCloud TopClouds[10];
public:
	void Clear(void);
	void Draw(void);
	void DrawCloud(const char *TexName, GLfloat *Pos);
	void NextWeather(void);
	void PreviousWeather(void);
	void GetSky_top(byte *color);
	void GetSky_bot(byte *color);
	void GetAmb(byte *color);
	void GetSunCore(byte *color);
	void GetSunCorona(byte *color);
	void GetWaterColor(byte *color);
	void GetTintColor(byte *color);
	void GetLowCloudsRGB(byte *color);
	void GetTopCloudRGB(byte *color);
	void GetBottomCloudRGB(byte *color);

	/* Not quite correct */
	GLfloat GetFarClp(void);
	GLfloat GetFogSt(void);
	GLfloat GetSunSz(void);
	GLfloat GetSprSz(void);

	void ReadWeatherData(void);
	int  GetWeatherSelector(void);
	void SetSunPos(const GLfloat x, const GLfloat y, const GLfloat z);
	void UpdateSun(void);
	void Init(void);
	void Dump(void);
	CSky(void);
	~CSky(void);
};

class CClock {
private:
	int Hour;
	int Minute;
public:
	int GetHour(void);
	int GetMinute(void);
	void SetHour(int h);
	void SetMinute(int m);
	void AdvanceOneMinute(void);
	void AdvanceOneHour(void);
	void TurnBackOneMinute(void);
	void TurnBackOneHour(void);
};

extern uint Game;
extern char progname[30];
extern bool UseTextures;
extern bool UseVertexColors;
extern bool UseFog;
extern bool UseTint;
extern bool UseTime;
extern bool DoStereo;
extern bool DoRadar;
extern uint Interior;
extern char *camposfile;
extern GLfloat WorldLimits[4];
extern GLfloat LODmultiplier;

extern CCamera Camera;
extern CDirectory Directory;
extern CDefinitionList DefinitionList;
extern CInstanceList InstanceList;
extern CFileLoader FileLoader;
extern CWater Water;
extern CSky Sky;
extern CClock Clock;
extern int nRadarFiles;
extern int nMilliseconds;

extern GLfloat planex;
extern GLfloat planey;
extern GLfloat planescale;
extern GLfloat lgtamb[4];

bool BindTexture(const char *texname);
bool BindTextureFromTxd(const char *TexName, const char *DictName);

char *GetWordUp(char *line, char *word);
char *GetQuotedWordUp(char *line, char *word);
char *GetFloat(char *line, float *number);
char *GetInt(char *line, void *number);
char *GetRGB(char *line, byte *color);
char *GetRGBA(char *line, byte *color);
char *ReadConfigLine(FILE *file, char *buffer);
void MixColors(RwUInt8 *result, RwUInt8 *col1,
               RwUInt8 *col2, int ratio, int all);

void StartOgl(int, char**);
void UpdateScene(int ms);
void RenderScene(void);
void RenderViewport(void);
void DrawWorld(int DrawTransp);
void DrawRadar(void);
bool isVisible(CWorldObj *Obj);
void TransformItem(Instance *Inst);
void DrawSprite(const char *DictName, const char *TexName, bool hasAlpha);
void Init(void);
void SwapBuffers(void);
int getWidth(void);
int getHeight(void);

void MakeScreenshot(void);

extern GLfloat lgtamb[4];

#endif
