#include <GL/glu.h>
#include <cmath>
#include "mapviewer.h"

void CCamera::CalcLookVector()
{
	GLfloat t1, t2;

	LookVector[0] = cos(Pitch);
	LookVector[1] = 0;
	LookVector[2] = sin(Pitch);

	t1 = (LookVector[0] * cos(Yaw)) - (LookVector[1] * sin(Yaw));
	t2 = (LookVector[1] * cos(Yaw)) + (LookVector[0] * sin(Yaw));
	LookVector[0] = t1;
	LookVector[1] = t2;
}

void CCamera::SetYaw(GLfloat y)
{
	Yaw = y;
	CalcLookVector();
}

GLfloat CCamera::GetYaw()
{
	return Yaw;
}

void CCamera::SetPitch(GLfloat p)
{
	Pitch = p;
	CalcLookVector();
}

void CCamera::SetPosition(GLfloat x, GLfloat y, GLfloat z)
{
	Position[0] = x;
	Position[1] = y;
	Position[2] = z;
}

void CCamera::TurnLeftRight(GLfloat ang)
{
	Yaw += ang;
	CalcLookVector();
}

void CCamera::TurnUpDown(GLfloat ang)
{
	if (Pitch + ang < M_PI/2 && Pitch + ang > -M_PI/2) {
		Pitch += ang;
		CalcLookVector();
	}
}

void CCamera::MoveForwardBackward(GLfloat dist)
{
	Position[0] += LookVector[0]*dist;
	Position[1] += LookVector[1]*dist;
	Position[2] += LookVector[2]*dist;
}

void CCamera::MoveUpDown(GLfloat dist)
{
	Position[2] += dist;
}

void CCamera::StrafeLeftRight(GLfloat dist)
{
	GLfloat direction[3];
	GLfloat up[] = { 0.0f, 0.0f, 1.0f };

	direction[0] = up[1]*LookVector[2] - up[2]*LookVector[1];
	direction[1] = up[2]*LookVector[0] - up[0]*LookVector[2];
	direction[2] = up[0]*LookVector[1] - up[1]*LookVector[0];

	Position[0] += direction[0]*dist;
	Position[1] += direction[1]*dist;
	Position[2] += direction[2]*dist;
}

void CCamera::Look()
{
	gluLookAt(Position[0], Position[1], Position[2],
	          Position[0]+LookVector[0], Position[1]+LookVector[1],
	          Position[2]+LookVector[2],
	          0.0f, 0.0f, 1.0f);
}

float CCamera::DistanceTo(float *vector)
{
	float a[3];
	a[0] = vector[0] - Position[0];
	a[1] = vector[1] - Position[1];
	a[2] = vector[2] - Position[2];
	return sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[3]);
}

void CCamera::ReadSavedPositions()
{
	int i;
	char Buffer[512];
	FILE *file;

	file = fopen(camposfile, "r");
	if (file != NULL) {
		for (i = 0; i < 12; i++) {
			if (ReadConfigLine(file, Buffer) == NULL)
				return;
			sscanf(Buffer, "%f %f %f %f %f", &CamPositions[i][0], 
			       &CamPositions[i][1], &CamPositions[i][2],
			       &CamPositions[i][3], &CamPositions[i][4]);
		}
		fclose(file);
	}
}

void CCamera::SavePosition(int PosNumber)
{
	int i;
	FILE *file;

	if ((file = fopen(camposfile, "w")) == NULL) {
		fprintf(stderr, "Couldn't save Camera Positions\n");
		fprintf(stderr, "%s\n", camposfile);
		return;
	}
	CamPositions[PosNumber][0] = Position[0];
	CamPositions[PosNumber][1] = Position[1];
	CamPositions[PosNumber][2] = Position[2];
	CamPositions[PosNumber][3] = Yaw;
	CamPositions[PosNumber][4] = Pitch;
	for (i = 0; i < 12; i++) {
		fprintf(file, "%f %f %f %f %f\n", CamPositions[i][0], 
			CamPositions[i][1], CamPositions[i][2],
			CamPositions[i][3], CamPositions[i][4]);
	}
}

void CCamera::LoadPosition(int PosNumber)
{
	Position[0] = CamPositions[PosNumber][0];
	Position[1] = CamPositions[PosNumber][1];
	Position[2] = CamPositions[PosNumber][2];
	Yaw = CamPositions[PosNumber][3];
	Pitch = CamPositions[PosNumber][4];
	CalcLookVector();
}

void CCamera::GetPosition(GLfloat *x, GLfloat *y, GLfloat *z)
{
	*x = Position[0];
	*y = Position[1];
	*z = Position[2];
}

CCamera::CCamera()
{
	Position[0] = 0;
	Position[1] = 0;
	Position[2] = 0;
	LookVector[0] = 1;
	LookVector[1] = 0;
	LookVector[2] = 0;
	Yaw = 0;
	Pitch = 0;
	memset(CamPositions, 0, 5*12*sizeof(GLfloat));
}

CCamera::~CCamera()
{
}
