#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <vector>
#include <GL/gl.h>
#include <GL/glu.h>

#include <rwcore.h>
#include <rpworld.h>

#ifndef DFFVIEWER_H
#define DFFVIEWER_H

#define GAMEPC  0x1
#define GAMEPS2 0x2
#define GAMEIII	0x4
#define GAMEVICECITY 0x8
#define GAMESANANDREAS 0x10

#define BACKPLANE 2000.0f
//#define BACKPLANE 10000.0f

using namespace std;

typedef unsigned int uint;
typedef unsigned char byte;
typedef byte color[4];

class CCamera {
private:
	GLfloat Position[3];
	GLfloat LookVector[3];
	GLfloat Yaw;
	GLfloat Pitch;
	GLfloat CamPositions[12][5];
	void CalcLookVector(void);
public:
	void SetYaw(GLfloat);
	GLfloat GetYaw(void);
	void SetPitch(GLfloat);
	void SetPosition(GLfloat, GLfloat, GLfloat);
	void TurnLeftRight(GLfloat);
	void TurnUpDown(GLfloat);
	void MoveForwardBackward(GLfloat);
	void MoveUpDown(GLfloat);
	void StrafeLeftRight(GLfloat);
	void Look(void);
	float DistanceTo(float *vector);
	void ReadSavedPositions(void);
	void SavePosition(int PosNumber);
	void LoadPosition(int PosNumber);
	void GetPosition(GLfloat *x, GLfloat *y, GLfloat *z);
	CCamera();
	~CCamera();
};

typedef struct DirEntry DirEntry;
struct DirEntry {
	char FileName[48]; /* file name used by GTA */
	uint Start;
	uint Size;
	char ContFile[48]; /* the actual file */
	bool Unloadable;

	/* Either Dff or Txd */
	void *Data;
	DirEntry *Parent;
};

class CDirectory {
private:
	vector<DirEntry> FileList;
public:
	void AppendFromDir(FILE *Dir, const char *ContFile);
	void Append(DirEntry&);
	void Append(const char *FileName, uint Start, uint Size,
	            const char *ContFile, void * Data);
	DirEntry *GetEntry(const char *FileName);
	DirEntry *GetEntry(const char *Name, const char *Ext);
	void Dump(void);
};

class CWorldObj {
public:
	int ID;
	char ModelName[25];
	char TextureName[25];
	char AnimName[25];
	int ObjCount;
	float *DrawDist;
	int Flags;
	int TimeOn;
	int TimeOff;
	bool isTimed;

	RpClump *Clump;
	RwTexDictionary *TexDict;

	DirEntry *DffEntry;
	DirEntry *TxdEntry;

	int RefCount;

	void SetClump(RpClump *);
	RpClump *GetClump(void);
	void TransformFrame(RwFrame *, RwInt32);
	void DrawGeometry(RpGeometry&, bool);
	void Draw(bool DoTransformation, bool DrawTransp);
};

/*
	void DrawAtomicDL(RwInt32 index);
	void CreateDisplayLists(void);
	void DestroyDisplayLists(void);
*/

class CObjectList {
private:
	vector<CWorldObj> ObjList;
public:
	void AppendFromFile(FILE *file);
	CWorldObj *GetObj(int ID);
	void Dump(void);
};

typedef struct Instance Instance;
struct Instance {
	int ID;
	char Name[64];
	int Interior;
	Vector3f PositionV;
	Vector3f ScaleV;
	Vector4f Rotation;
	int LOD;

	CWorldObj *Obj;

	Instance *Parent;
	Instance *Child;
	int Index;

	bool isLOD;
};

class CInstanceList {
private:
	vector<Instance> InstList;

	int Index;
public:
	void AppendFromFile(FILE *file, char *name);
	void AppendFromPlainFile(FILE *file);
	void AppendFromBinFile(FILE *file);
	Instance *GetFirst(void);
	int GetLength(void);
	Instance *GetInst(int index);
	Instance *GetLOD(Instance *inst);
	Instance *GetCorrectInst(Instance *inst);
	void Dump(void);
};


class CFileLoader {
public:
	RpClump *LoadClump(DirEntry *);
	RwTexDictionary *LoadTexDict(DirEntry *);
	void FreeRwTextures(RwTexDictionary *);
	void LoadGLTextures(RwTexDictionary *);
	bool LoadDff(CWorldObj *obj);
	bool LoadTxd(CWorldObj *obj);
	void UnloadWorldObj(CWorldObj *obj);
	void LoadBinIpl(DirEntry *Entry);
};

class CWater {
private:
	byte indiceshi[128*128];
	byte indiceslo[64*64];
	float palette[48]; /* not sure */
	void DrawPatch(float x, float y, float patchsize, float height);
public:
	GLuint TexID;
	void Draw(void);
	void ReadWaterData(void);
};

typedef struct Weather Weather;
struct Weather {
	color Amb;
	color Amb_Obj;
	color Amb_Bl;
	color Amb_Obj_bl;
	color Dir;
	color Sky_top;
	color Sky_bot;
	color SunCore;
	color SunCorona;
	float SunSz;
	float SprSz;
	float SprBght;
	int Shdw;
	int LightShd;
	int TreeShd;
	int PoleShd;
	float FarClp;
	float FogSt;
	float LightOnGround;
	color LowCloudsRGB;
	color TopCloudRGB;
	color BottomCloudRGB;

	/* GTA 3 multiplied with 0.6 (AK-73) */
	color Tint;

	color BlurRGB;
	color WaterRGBA;

	color RGBA1;
	color RGBA2;
	float CloudAlpha;
};

class CSky {
private:
	vector<Weather> WeatherList;
	int numWeather;
	int WeatherSelector;
public:
	void Draw(void);
	void ReadWeatherData(void);
	void NextWeather(void);
	void PreviousWeather(void);
	void GetSky_top(byte *color);
	void GetSky_bot(byte *color);
	void GetWaterColor(byte *color);
	void GetTintColor(byte *color);
	int GetWeatherSelector(void);
	void Dump(void);
	CSky(void);
	~CSky(void);
};

class CClock {
private:
	int Hour;
	int Minute;
public:
	int GetHour(void);
	int GetMinute(void);
	void SetHour(int h);
	void SetMinute(int m);
	void AdvanceOneMinute(void);
	void AdvanceOneHour(void);
	void TurnBackOneMinute(void);
	void TurnBackOneHour(void);
};

extern uint Game;
extern char progname[30];
extern bool UseTextures;
extern bool UseVertexColors;
extern bool UseFog;
extern bool UseTint;
extern uint Interior;
extern char *camposfile;
extern GLfloat FOV;

extern CCamera Camera;
extern CDirectory Directory;
extern CObjectList ObjectList;
extern CInstanceList InstanceList;
extern CFileLoader FileLoader;
extern CWater Water;
extern CSky Sky;
extern CClock Clock;


extern GLfloat planex;
extern GLfloat planey;
extern GLfloat planescale;
extern GLfloat lgtamb[4];


bool BindTexture(const char *texname);

char *GetWordUp(char *line, char *word);
char *GetFloat(char *line, float *number);
char *GetInt(char *line, void *number);
char *GetRGB(char *line, byte *color);
char *GetRGBA(char *line, byte *color);
char *ReadConfigLine(FILE *file, char *buffer);
void MixColors(RwUInt8 *result, RwUInt8 *col1,
               RwUInt8 *col2, int ratio, int all);

void StartOgl(int, char**);
void RenderScene(void);
void Init(void);
void DeInit(void);
void SwapBuffers(void);
int getWidth(void);
int getHeight(void);

#endif
